#!/usr/bin/env python

import gcc

import os, sys

from PyQt4.QtCore import *
from PyQt4.QtGui import *

# --------------------------------------------------------------------------

def findIcon (icon_name):
    icon = QIcon.fromTheme (icon_name)
    return icon

class TreeItem (QTreeWidgetItem):
    def __init__ (self, parent, text):
        QTreeWidgetItem.__init__ (self, parent)
        self.setText (0, text)

    def addIcon (self, icon_name):
        self.setIcon (0, findIcon (icon_name))

class TreeView (QTreeWidget):
    def __init__ (self, parent=None):
        QTreeWidget.__init__ (self, parent)
        self.setColumnCount (1)
        self.header ().hide ()
        # self.setRootIsDecorated (True)
        # self.setSelectionMode (QAbstractItemView.ExtendedSelection)
        self.setEditTriggers (QAbstractItemView.SelectedClicked)
        self.setIndentation (8)

# --------------------------------------------------------------------------

class MainWin (QMainWindow):

    def __init__ (self, parent=None):
        QMainWindow.__init__ (self, parent)

        self.editors = { }
        self.types = { }

        self.setStatusBar (QStatusBar ())
        self.resize (640, 480)

        self.split = QSplitter ()
        self.setCentralWidget (self.split)

        self.treeView = TreeView ()
        self.split.addWidget (self.treeView)

        self.rightTabs = QTabWidget ()
        self.rightTabs.tabBar().setFocusPolicy (Qt.NoFocus)
        self.rightTabs.setTabPosition (QTabWidget.South)
        self.split.addWidget (self.rightTabs)

        fileMenu = self.menuBar().addMenu ("&File")
        act = QAction ("&Quit", self)
        act.triggered.connect (self.close)
        fileMenu.addAction (act)

        self.connect (self.treeView,
                       SIGNAL ("itemClicked (QTreeWidgetItem *, int)"),
                       self.treeView_itemClicked)

    def treeView_itemClicked (self, tree_node, column) :
        txt = ""
        if hasattr (tree_node, "location") :
           loc = tree_node.location
           txt = loc.file + ", line: " + str (loc.line) + ", column: " + str (loc.column)
           self.openSource (loc.file, loc.line, loc.column, loc.line, loc.column)
        self.statusBar().showMessage (txt)

    def openSource (self, fileName, line, col, stop_line, stop_col):
        if fileName not in self.editors :

           name = QString  (os.path.basename (fileName))

           view = QTextEdit ()
           view.setWordWrapMode (False)

           self.rightTabs.addTab (view, name)
           self.editors [fileName] = view

           text = open (fileName).read ()
           view.setText (text)

        edit = self.editors [fileName]
        self.rightTabs.setCurrentWidget (edit)

        cursor = edit.textCursor ()
        cursor.movePosition (QTextCursor.Start)
        cursor.movePosition (QTextCursor.Down, QTextCursor.MoveAnchor, line-1)

        if True :
           cursor.movePosition (QTextCursor.StartOfLine, QTextCursor.MoveAnchor)
           cursor.movePosition (QTextCursor.EndOfLine, QTextCursor.KeepAnchor)
        else :
           cursor.movePosition (QTextCursor.Right, QTextCursor.MoveAnchor, col-1)

           if stop_line == line :
              cursor.movePosition (QTextCursor.Right, QTextCursor.KeepAnchor, stop_col-col)
           else :
              cursor.movePosition (QTextCursor.Down, QTextCursor.KeepAnchor, stop_line-line)
              cursor.movePosition (QTextCursor.StartOfLine, QTextCursor.KeepAnchor)
              cursor.movePosition (QTextCursor.Right, QTextCursor.KeepAnchor, stop_col)

        edit.setTextCursor (cursor)

    def addBranch (self, above, node, level):

        if level > 10 :
           return

        # if hasattr (node, 'is_artificial'):
        #    if node.is_artificial :
        #          return

        if hasattr (node, 'is_builtin'):
           if node.is_builtin :
              if level != 1 : # enable global namespace
                 return

        txt = ""
        if isinstance (node, gcc.FunctionDecl) :
           txt = node.name + " : " + str (node.type) + " : " + node.__class__.__name__
        elif isinstance (node, gcc.Declaration) :
           txt = str (node.name) + " : " + str (node.type) + " : " + node.__class__.__name__

        unknown = False
        if txt == "" :
           txt = str (node) + " : " + node.__class__.__name__
           unknown = True

        item = TreeItem (above, txt)

        if isinstance (node, gcc.FunctionDecl) :
           item.addIcon ("code-function")
        if isinstance (node, gcc.TypeDecl) :
           item.addIcon ("code-class")
        if isinstance (node, gcc.FieldDecl) :
           item.addIcon ("code-variable")
        if isinstance (node, gcc.VarDecl) :
           item.addIcon ("code-variable")
        if isinstance (node, gcc.ConstDecl) :
           item.addIcon ("code-variable")
        if isinstance (node, gcc.NamespaceDecl) :
           item.addIcon ("code-block")

        if isinstance (node, gcc.Declaration) :
           item.location = node.location


        if isinstance (node, gcc.NamespaceDecl) :
           for t in node.namespaces :
               self.addBranch (item, t, level+1)
           for t in node.declarations :
               self.addBranch (item, t, level+1)
        elif isinstance (node, gcc.FunctionDecl) :
           for t in node.arguments :
               self.addBranch (item, t, level+1)
           if node.result :
               self.addBranch (item, node.result, level+1)
           #if node.type :
              #self.addBranch (item, node.type, level+1)
        #elif isinstance (node, gcc.VarDecl) or isinstance (node, gcc.ParmDecl):
           #if node.type :
              #self.addBranch (item, node.type, level+1)
        # elif isinstance (node, gcc.EnumeralType) :
        #    for t in node.values :
        #        self.addBranch (item, t, level+1)
        elif isinstance (node, gcc.RecordType) :
           for t in node.fields :
               self.addBranch (item, t, level+1)
           for t in node.methods :
               self.addBranch (item, t, level+1)
           # if node.type :
           #     self.addBranch (item, node.type, level+1)

        if isinstance (node, gcc.TypeDecl) or unknown and hasattr (node, "type"):
           id = str (node.type)
           if id not in self.types or unknown :
              self.types [id] = True
              self.addBranch (item, node.type, level+1)

           if hasattr (node, "dereference") :
              sub_item = TreeItem (item, "dereference")
              self.addBranch (sub_item, node.dereference, level+1)

        #if showFields :
           #text = str (dir (node))
           #for s in text.split (',') :
               #TreeItem (item, s)

# --------------------------------------------------------------------------

def on_pass_execution (p, fn):
    if p.name == '*warn_function_return':
        # For this pass, "fn" will be an instance of gcc.Function:
        print ('fn: %r' % fn)
        print ('fn.decl.name: %r' % fn.decl.name)

        # fn.decl is an instance of gcc.FunctionDecl:
        print ('return type: %r' % str (fn.decl.type.type))
        print ('argument types: %r' % [str (t) for t in fn.decl.type.argument_types])

    if p.name == '*free_lang_data':
        global win
        win.addBranch (win.treeView, gcc.get_global_namespace (), 1)

    if p.name == '*clean_state':
        global app
        app.exec_ ()

# --------------------------------------------------------------------------

QIcon.setThemeName ("oxygen")
app = QApplication ([""])
win = None
win = MainWin ()
win.show ()
gcc.register_callback (gcc.PLUGIN_PASS_EXECUTION, on_pass_execution)
# NO app.exec_ ()

# --------------------------------------------------------------------------

# kate: indent-width 1; show-tabs true; replace-tabs true; remove-trailing-spaces all
